#define _GNU_SOURCE

#include <pwd.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mount.h>
#include <sys/syslog.h>
#include <unistd.h>

#include <security/pam_modules.h>
#include <security/pam_ext.h>

#define TMPDIR "/tmp"

int pam_sm_open_session(pam_handle_t *pamh, int flags,
			int argc, const char **argv)
{
	if (geteuid() != 0)
		return PAM_SESSION_ERR;

	const void *item;
	int status = pam_get_item(pamh, PAM_USER, &item);
	if (PAM_SUCCESS != status)
		return status;

	const char *const user = item;
	struct passwd *pw = getpwnam(user);
	if (!pw)
		return PAM_USER_UNKNOWN;
	memset(pw->pw_passwd, 0, strlen(pw->pw_passwd));

	if (pw->pw_uid == 0)
		return PAM_IGNORE;

	if (unshare(CLONE_NEWNS)) {
		pam_syslog(pamh, LOG_ERR, "unshare CLONE_NEWNS: %m");
		return PAM_SESSION_ERR;
	}

	if (mount("/", "/", NULL, MS_SLAVE|MS_REC, NULL)) {
		pam_syslog(pamh, LOG_ERR, "mount / MS_SLAVE|MS_REC: %m");
		return PAM_SESSION_ERR;
	}

	char *opts;
	if (asprintf(&opts, "mode=0700,uid=%u,gid=%u,size=64g",
		     pw->pw_uid, pw->pw_gid) < 0)
		return PAM_BUF_ERR;

	if (mount("tmpfs", TMPDIR, "tmpfs", MS_MGC_VAL|MS_NOSUID|MS_NODEV,
		  opts)) {
		pam_syslog(pamh, LOG_ERR, "mount: " TMPDIR ": %m");
		free(opts);
		return PAM_SESSION_ERR;
	}

	free(opts);

	status = pam_putenv(pamh, "TMPDIR=" TMPDIR);
	if (PAM_SUCCESS == status)
		status = pam_putenv(pamh, "TMP=" TMPDIR);

	return status;
}

int pam_sm_close_session(pam_handle_t *pamh, int flags,
			 int argc, const char **argv)
{
	return PAM_SUCCESS;
}
